@@ -1,11 +1,6 @@
-#!usr/bin/python
-# -*- coding: utf-8 -*-
-
-"""
-GradCAM
-"""
-
 import torch
+from torch import Tensor
+from typing import Optional, Tuple
 
 from .cam import _CAM
 
@@ -16,29 +11,34 @@ class _GradCAM(_CAM):
     """Implements a gradient-based class activation map extractor
 
     Args:
-        model (torch.nn.Module): input model
-        conv_layer (str): name of the last convolutional layer
+        model: input model
+        target_layer: name of the target layer
+        input_shape: shape of the expected input tensor excluding the batch dimension
     """
 
-    hook_a, hook_g = None, None
-    hook_handles = []
+    def __init__(
+        self,
+        model: torch.nn.Module,
+        target_layer: Optional[str] = None,
+        input_shape: Tuple[int, ...] = (3, 224, 224),
+    ) -> None:
 
-    def __init__(self, model, conv_layer):
-
-        super().__init__(model, conv_layer)
+        super().__init__(model, target_layer, input_shape)
+        # Init hook
+        self.hook_g: Optional[Tensor] = None
         # Ensure ReLU is applied before normalization
         self._relu = True
         # Model output is used by the extractor
         self._score_used = True
         # Backward hook
-        self.hook_handles.append(self.model._modules.get(conv_layer).register_backward_hook(self._hook_g))
+        self.hook_handles.append(self.submodule_dict[self.target_layer].register_backward_hook(self._hook_g))
 
-    def _hook_g(self, module, input, output):
+    def _hook_g(self, module: torch.nn.Module, input: Tensor, output: Tensor) -> None:
         """Gradient hook"""
         if self._hooks_enabled:
             self.hook_g = output[0].data
 
-    def _backprop(self, scores, class_idx):
+    def _backprop(self, scores: Tensor, class_idx: int) -> None:
         """Backpropagate the loss for a specific output class"""
 
         if self.hook_a is None:
@@ -69,7 +69,7 @@ class GradCAM(_GradCAM):
         w_k^{(c)} = \\frac{1}{H \\cdot W} \\sum\\limits_{i=1}^H \\sum\\limits_{j=1}^W
         \\frac{\\partial Y^{(c)}}{\\partial A_k(i, j)}
 
-    where :math:`A_k(x, y)` is the activation of node :math:`k` in the last convolutional layer of the model at
+    where :math:`A_k(x, y)` is the activation of node :math:`k` in the target layer of the model at
     position :math:`(x, y)`,
     and :math:`Y^{(c)}` is the model output score for class :math:`c` before softmax.
 
@@ -78,27 +78,23 @@ class GradCAM(_GradCAM):
         >>> from torchcam.cams import GradCAM
         >>> model = resnet18(pretrained=True).eval()
         >>> cam = GradCAM(model, 'layer4')
-        >>> with torch.no_grad(): scores = model(input_tensor)
+        >>> scores = model(input_tensor)
         >>> cam(class_idx=100, scores=scores)
 
     Args:
-        model (torch.nn.Module): input model
-        conv_layer (str): name of the last convolutional layer
+        model: input model
+        target_layer: name of the target layer
+        input_shape: shape of the expected input tensor excluding the batch dimension
     """
 
-    hook_a, hook_g = None, None
-
-    def __init__(self, model, conv_layer):
-
-        super().__init__(model, conv_layer)
-
-    def _get_weights(self, class_idx, scores):
+    def _get_weights(self, class_idx: int, scores: Tensor) -> Tensor:  # type: ignore[override]
         """Computes the weight coefficients of the hooked activation maps"""
 
+        self.hook_g: Tensor
         # Backpropagate
         self._backprop(scores, class_idx)
         # Global average pool the gradients over spatial dimensions
-        return self.hook_g.squeeze(0).mean(axis=(1, 2))
+        return self.hook_g.squeeze(0).mean(dim=(1, 2))
 
 
 class GradCAMpp(_GradCAM):
@@ -116,7 +112,7 @@ class GradCAMpp(_GradCAM):
         w_k^{(c)} = \\sum\\limits_{i=1}^H \\sum\\limits_{j=1}^W \\alpha_k^{(c)}(i, j) \\cdot
         ReLU\\Big(\\frac{\\partial Y^{(c)}}{\\partial A_k(i, j)}\\Big)
 
-    where :math:`A_k(x, y)` is the activation of node :math:`k` in the last convolutional layer of the model at
+    where :math:`A_k(x, y)` is the activation of node :math:`k` in the target layer of the model at
     position :math:`(x, y)`,
     :math:`Y^{(c)}` is the model output score for class :math:`c` before softmax,
     and :math:`\\alpha_k^{(c)}(i, j)` being defined as:
@@ -134,32 +130,28 @@ class GradCAMpp(_GradCAM):
         >>> from torchcam.cams import GradCAMpp
         >>> model = resnet18(pretrained=True).eval()
         >>> cam = GradCAMpp(model, 'layer4')
-        >>> with torch.no_grad(): scores = model(input_tensor)
+        >>> scores = model(input_tensor)
         >>> cam(class_idx=100, scores=scores)
 
     Args:
-        model (torch.nn.Module): input model
-        conv_layer (str): name of the last convolutional layer
+        model: input model
+        target_layer: name of the target layer
+        input_shape: shape of the expected input tensor excluding the batch dimension
     """
 
-    hook_a, hook_g = None, None
-
-    def __init__(self, model, conv_layer):
-
-        super().__init__(model, conv_layer)
-
-    def _get_weights(self, class_idx, scores):
+    def _get_weights(self, class_idx: int, scores: Tensor) -> Tensor:  # type: ignore[override]
         """Computes the weight coefficients of the hooked activation maps"""
 
+        self.hook_g: Tensor
         # Backpropagate
         self._backprop(scores, class_idx)
         # Alpha coefficient for each pixel
         grad_2 = self.hook_g.pow(2)
-        grad_3 = self.hook_g.pow(3)
-        alpha = grad_2 / (2 * grad_2 + (grad_3 * self.hook_a).sum(axis=(2, 3), keepdims=True))
+        grad_3 = grad_2 * self.hook_g
+        alpha = grad_2 / (2 * grad_2 + (grad_3 * self.hook_a).sum(dim=(2, 3), keepdim=True))
 
         # Apply pixel coefficient in each weight
-        return alpha.squeeze_(0).mul_(torch.relu(self.hook_g.squeeze(0))).sum(axis=(1, 2))
+        return alpha.squeeze_(0).mul_(torch.relu(self.hook_g.squeeze(0))).sum(dim=(1, 2))
 
 
 class SmoothGradCAMpp(_GradCAM):
@@ -178,7 +170,7 @@ class SmoothGradCAMpp(_GradCAM):
         w_k^{(c)} = \\sum\\limits_{i=1}^H \\sum\\limits_{j=1}^W \\alpha_k^{(c)}(i, j) \\cdot
         ReLU\\Big(\\frac{\\partial Y^{(c)}}{\\partial A_k(i, j)}\\Big)
 
-    where :math:`A_k(x, y)` is the activation of node :math:`k` in the last convolutional layer of the model at
+    where :math:`A_k(x, y)` is the activation of node :math:`k` in the target layer of the model at
     position :math:`(x, y)`,
     :math:`Y^{(c)}` is the model output score for class :math:`c` before softmax,
     and :math:`\\alpha_k^{(c)}(i, j)` being defined as:
@@ -203,26 +195,33 @@ class SmoothGradCAMpp(_GradCAM):
         >>> from torchvision.models import resnet18
         >>> from torchcam.cams import SmoothGradCAMpp
         >>> model = resnet18(pretrained=True).eval()
-        >>> cam = SmoothGradCAMpp(model, 'layer4', 'conv1')
-        >>> with torch.no_grad(): scores = model(input_tensor)
+        >>> cam = SmoothGradCAMpp(model, 'layer4')
+        >>> scores = model(input_tensor)
         >>> cam(class_idx=100)
 
     Args:
-        model (torch.nn.Module): input model
-        conv_layer (str): name of the last convolutional layer
+        model: input model
+        target_layer: name of the target layer
+        num_samples: number of samples to use for smoothing
+        std: standard deviation of the noise
+        input_shape: shape of the expected input tensor excluding the batch dimension
     """
 
-    hook_a, hook_g = None, None
-    hook_handles = []
+    def __init__(
+        self,
+        model: torch.nn.Module,
+        target_layer: Optional[str] = None,
+        num_samples: int = 4,
+        std: float = 0.3,
+        input_shape: Tuple[int, ...] = (3, 224, 224),
+    ) -> None:
 
-    def __init__(self, model, conv_layer, first_layer, num_samples=4, std=0.3):
-
-        super().__init__(model, conv_layer)
+        super().__init__(model, target_layer, input_shape)
         # Model scores is not used by the extractor
         self._score_used = False
 
         # Input hook
-        self.hook_handles.append(self.model._modules.get(first_layer).register_forward_pre_hook(self._store_input))
+        self.hook_handles.append(model.register_forward_pre_hook(self._store_input))
         # Noise distribution
         self.num_samples = num_samples
         self.std = std
@@ -230,15 +229,17 @@ class SmoothGradCAMpp(_GradCAM):
         # Specific input hook updater
         self._ihook_enabled = True
 
-    def _store_input(self, module, input):
+    def _store_input(self, module: torch.nn.Module, input: Tensor) -> None:
         """Store model input tensor"""
 
         if self._ihook_enabled:
             self._input = input[0].data.clone()
 
-    def _get_weights(self, class_idx, scores=None):
+    def _get_weights(self, class_idx: int, scores: Optional[Tensor] = None) -> Tensor:
         """Computes the weight coefficients of the hooked activation maps"""
 
+        self.hook_a: Tensor
+        self.hook_g: Tensor
         # Disable input update
         self._ihook_enabled = False
         # Keep initial activation
@@ -266,10 +267,10 @@ class SmoothGradCAMpp(_GradCAM):
         grad_3.div_(self.num_samples)
 
         # Alpha coefficient for each pixel
-        alpha = grad_2 / (2 * grad_2 + (grad_3 * init_fmap).sum(axis=(2, 3), keepdims=True))
+        alpha = grad_2 / (2 * grad_2 + (grad_3 * init_fmap).sum(dim=(2, 3), keepdim=True))
 
         # Apply pixel coefficient in each weight
-        return alpha.squeeze_(0).mul_(torch.relu(self.hook_g.squeeze(0))).sum(axis=(1, 2))
+        return alpha.squeeze_(0).mul_(torch.relu(self.hook_g.squeeze(0))).sum(dim=(1, 2))
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return f"{self.__class__.__name__}(num_samples={self.num_samples}, std={self.std})"
